/*!
 * Copyright (C) 2006-2007 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * \file config.c
 * \brief Radio configuration.
 *
 * \verbatim
 *
 * $Log$
 *
 * \endverbatim
 */

#include <cfg/os.h>
#include <cfg/clock.h>
#include <dev/board.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <io.h>

#include <sys/version.h>
#include <sys/confnet.h>
#include <sys/atom.h>
#include <sys/heap.h>
#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/event.h>
#include <sys/socket.h>

#include <hxmp3/mp3dec.h>

#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <net/route.h>
#include <pro/dhcp.h>

#include "config.h"

HTTP_PROXY proxy;
STATIONCONF *station;
RADIOCONTROL radio;

/*!
 * \brief Configure a station list entry.
 *
 * \param idx Index of the entry. Use MAXNUM_STATIONS to append an entry.
 * \param url URL of the station.
 *
 * \return 0 on success, -1 otherwise.
 */
int ConfigStation(u_char idx, CONST char *symbol, CONST char * url)
{
    u_long ip;
    u_short port = 80;
    char *buf;
    char *cp;

    if (idx >= MAXNUM_STATIONS) {
        /* Appending an entry. Find a free slot. */
        idx = 0;
        while (idx < MAXNUM_STATIONS && station[idx].rs_port)
            idx++;
    }
    if (idx >= MAXNUM_STATIONS) {
        return -1;
    }

    buf = malloc(strlen(url) + 1);

    /* Extract IP address. */
    cp = buf;
    while (*url && *url != '/' && *url != ':')
        *cp++ = *url++;
    *cp = 0;
    if ((int) (ip = inet_addr(buf)) == -1)
        idx = -1;
    else {
        /* Extract URL path. */
        cp = buf;
        if (*url == '/') {
            url++;
            while (*url && *url != ':')
                *cp++ = *url++;
        }
        *cp = 0;

        /* Extract port. */
        if (*url == ':')
            port = atoi(url + 1);

        if (port) {
            station[idx].rs_ip = ip;
            station[idx].rs_port = port;
            if (*buf) {
                station[idx].rs_url = malloc(strlen(buf) + 1);
                strcpy(station[idx].rs_url, buf);
            }
        }
    }
    free(buf);

    return idx;
}

/*
 * Allocate heap memory for configuration structure.
 */
static void ConfigCreate(void)
{
    u_char idx;

    if (station == 0)
        station = malloc(MAXNUM_STATIONS * sizeof(STATIONCONF));
    else {
        /* Release all memory occupied by the current configuration. */
        for (idx = 0; idx < MAXNUM_STATIONS; idx++) {
            if (station[idx].rs_port == 0)
                break;
            if (station[idx].rs_url)
                free(station[idx].rs_url);
        }
    }
    memset(station, 0, MAXNUM_STATIONS * sizeof(STATIONCONF));
}

/*!
 * \brief Reset configuration.
 */
void ConfigResetFactory(void)
{
    ConfigCreate();

    /* Set HTTP proxy. */
#ifdef MY_PROXY_IP
    proxy.proxy_ip = inet_addr(MY_PROXY_IP);
    proxy.proxy_port = MY_PROXY_PORT;
#endif

    /* Initial radio control settings. */
    radio.rc_rstation = 2;
    radio.rc_rvolume = 223;

    /* 
     * Add pre-configured radio stations. 
     */
#if defined(AT91SAM7X_EK)
    /* Low bandwidth. */
    ConfigStation(MAXNUM_STATIONS, "SmoothJazz", "212.23.5.99:8004"); //16000 mono
    ConfigStation(MAXNUM_STATIONS, "Classical", "206.98.167.99:8406"); //16000 mono
    ConfigStation(MAXNUM_STATIONS, "SmoothJazz", "212.23.5.99:8004"); //16000 mono
    ConfigStation(MAXNUM_STATIONS, "Classical", "206.98.167.99:8406"); //16000 mono
    ConfigStation(MAXNUM_STATIONS, "SmoothJazz", "212.23.5.99:8004"); //16000 mono
    ConfigStation(MAXNUM_STATIONS, "Classical", "206.98.167.99:8406"); //16000 mono
    
#else
#if 0
    ConfigStation(MAXNUM_STATIONS, "egnite", "192.168.192.223:8000");
    ConfigStation(MAXNUM_STATIONS, "egnite", "192.168.192.223:8000");
#else
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "88.191.22.199:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "194.158.114.68:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "194.158.114.67:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "193.251.154.243:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "213.186.56.206:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "194.158.114.66:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "213.186.41.160:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "88.191.11.19:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "62.4.21.91:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "212.129.62.140:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "194.117.194.66:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "213.186.37.124:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "80.65.234.120:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "213.161.201.21:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "213.186.60.54:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "193.251.154.242:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "83.143.18.76:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "193.17.192.35:8000");
    ConfigStation(MAXNUM_STATIONS, "FREQUENCE3", "193.222.128.135:8000");

    ConfigStation(MAXNUM_STATIONS, "Germany", "88.84.142.48:8500"); //320kbps 44100 stereo
    ConfigStation(MAXNUM_STATIONS, "Plusz", "80.95.64.86:8002"); //256kbps 44100 stereo
    ConfigStation(MAXNUM_STATIONS, "PsychedeliC", "80.237.155.22:8016"); //256kbps 44100 stereo
    ConfigStation(MAXNUM_STATIONS, "PsychedeliC", "85.21.79.9:8000"); //256kbps 44100 stereo
    ConfigStation(MAXNUM_STATIONS, "PsychedeliC", "85.21.79.5:8000"); //256kbps 44100 stereo
    ConfigStation(MAXNUM_STATIONS, "SKY.FM-Jazz", "207.200.96.227:80/stream/1030");
    ConfigStation(MAXNUM_STATIONS, "SKY.FM-Jazz", "scfire-chi0l-1.stream.aol.com:80/stream/1010");
    ConfigStation(MAXNUM_STATIONS, "KPLU Jazz", "205.234.238.46:80"); //24000 mono
    ConfigStation(MAXNUM_STATIONS, "Oldies", "205.188.215.230:8000"); //44100 

    ConfigStation(MAXNUM_STATIONS, "WGBH Classic", "64.71.145.106:8004"); //44100 stereo with intro
    ConfigStation(MAXNUM_STATIONS, "Stellar", "194.79.31.249:7200"); //16000 stereo
    ConfigStation(MAXNUM_STATIONS, "SKY.FM Classic", "206.98.167.99:8406"); //16000 mono
    ConfigStation(MAXNUM_STATIONS, "SKY.FM Classic", "128.177.3.80:4078"); //16000 mono
    ConfigStation(MAXNUM_STATIONS, "SKY.FM Classic", "66.90.113.135:8010"); //16000 mono
    ConfigStation(MAXNUM_STATIONS, "KPLU Jazz", "205.234.238.46:80"); //24000 mono

    ConfigStation(MAXNUM_STATIONS, "AlexJones", "64.27.20.2:80");
    ConfigStation(MAXNUM_STATIONS, "AlexJones", "208.100.15.180:80");

    ConfigStation(MAXNUM_STATIONS, "TEST1", "64.236.34.97:80/stream/1005");
    ConfigStation(MAXNUM_STATIONS, "TEST2", "207.200.96.230:8040");
    ConfigStation(MAXNUM_STATIONS, "TEST3", "194.158.114.68:8000");
    ConfigStation(MAXNUM_STATIONS, "TEST4", "88.191.34.99:7020");
#endif
#endif
}

#if 0
typedef enum {
    CFGMN_MAIN = 0,
    CFGMN_CONFIG,
    CFGMN_NETWORK,
    CFGMN_LOCAL,
    CFGMN_LOCAL_IP,
    CFGMN_LOCAL_MASK,
    CFGMN_EXTERN,
    CFGMN_GATE,
    CFGMN_PROXY,
    CFGMN_PROXY_IP,
    CFGMN_PROXY_PORT,
    CFGMN_RADIO,
    CFGMN_VOLUME,
    CFGMN_STATION,
    CFGMN_NEW,
    CFGMN_NEW_IP,
    CFGMN_NEW_PORT,
    CFGMN_EDIT,
    CFGMN_SAVE,
    CFGMN_NO,
    CFGMN_YES,
    CFG_FUNCTION
} CFG_MENU_NUM;

typedef struct _CFG_MENU_ITEM {
    char *cfgm_title;
    int cfgm_btn[3];
} CFG_MENU_ITEM;

CFG_MENU_ITEM cfg_menu[] = {
    /*  0           */ { "Internet Radio", { CFGMN_CONFIG, CFGMN_SAVE, CFGMN_MAIN } },
    /*    1         */ { "Config"  , { CFGMN_NETWORK, CFGMN_RADIO, CFGMN_MAIN } },
    /*      2       */ { "Network" , { CFGMN_LOCAL,  CFGMN_EXTERN, CFGMN_CONFIG } },
    /*        3     */ { "Local"   , { CFGMN_LOCAL_IP, CFGMN_LOCAL_MASK, CFGMN_NETWORK } },
    /*          4   */ { "Address" , { CFG_FUNCTION, CFGMN_LOCAL_IP, CFGMN_LOCAL } },
    /*          5   */ { "Mask"    , { CFG_FUNCTION, CFGMN_LOCAL_MASK, CFGMN_LOCAL } },
    /*        6     */ { "Extern"  , { CFGMN_GATE, CFGMN_PROXY, CFGMN_NETWORK } },
    /*          7   */ { "Gate"    , { CFG_FUNCTION, CFGMN_GATE, CFGMN_EXTERN } },
    /*          8   */ { "Proxy"   , { CFGMN_PROXY_IP, CFGMN_PROXY_PORT, CFGMN_EXTERN } },
    /*            9 */ { "IP"      , { CFG_FUNCTION, CFGMN_PROXY_IP, CFGMN_PROXY } },
    /*           10 */ { "Port"    , { CFG_FUNCTION, CFGMN_PROXY_PORT, CFGMN_PROXY } },
    /*     11       */ { "Radio"   , { CFGMN_VOLUME, CFGMN_STATION, CFGMN_CONFIG } },
    /*       12     */ { "Volume"  , { CFG_FUNCTION, CFGMN_VOLUME, CFGMN_RADIO } },
    /*       13     */ { "Station" , { CFGMN_NEW, CFGMN_EDIT, CFGMN_RADIO } },
    /*         14   */ { "New"     , { CFGMN_NEW_IP, CFGMN_NEW_PORT, CFGMN_STATION } },
    /*           15 */ { "IP"      , { CFG_FUNCTION, CFGMN_NEW_IP, CFGMN_NEW } },
    /*           16 */ { "Port"    , { CFG_FUNCTION, CFGMN_NEW_PORT, CFGMN_NEW } },
    /*         17   */ { "Edit"    , { CFG_FUNCTION, CFGMN_EDIT, CFGMN_STATION } },
    /*   18         */ { "Save"    , { CFGMN_NO,  CFGMN_YES,  CFGMN_MAIN } },
    /*     19       */ { "NO"      , { CFGMN_MAIN,  CFGMN_MAIN,  CFGMN_MAIN } },
    /*     20       */ { "YES"     , { CFG_FUNCTION, CFGMN_YES, CFGMN_MAIN } }
};

THREAD(ConfigDialogThread, arg)
{
    u_long baud = 115200;
    FILE *uart;
    int cmd;
    int cpos = 0;

    /*
     * Register and open the UART used for basic configuration dialogs.
     */
    NutRegisterDevice(&DEV_UART0, 0, 0);
    uart = fopen(DEV_UART0_NAME, "r+");
    _ioctl(_fileno(uart), UART_SETSPEED, &baud);

    for (;;) {
        fprintf(uart, "\n%16s\n%8s%-8s", 
            cfg_menu[cpos].cfgm_title, 
            cfg_menu[cpos].cfgm_btn[0] ? cfg_menu[cfg_menu[cpos].cfgm_btn[0]].cfgm_title : "Back",
            cfg_menu[cpos].cfgm_btn[1] ? cfg_menu[cfg_menu[cpos].cfgm_btn[1]].cfgm_title : "Back");
        fflush(uart);
        cmd = fgetc(uart);
        if (cmd >= '1' && cmd <= '3') {
            cpos = cfg_menu[cpos].cfgm_btn[cmd -'1'];
        }
        if (cfg_menu[cpos].cfgm_btn[0] == CFG_FUNCTION) {
            switch(cfg_menu[cpos].cfgm_btn[1]) {
            case CFGMN_LOCAL_IP:
                //ConfigMenuIpAddr(&confnet.cdn_ip_addr);
                break;
            case CFGMN_LOCAL_MASK:
                //ConfigMenuIpAddr(&confnet.cdn_ip_mask);
                break;
            case CFGMN_GATE:
                //ConfigMenuIpAddr(&confnet.cdn_gateway);
                break;
            case CFGMN_PROXY_IP:
                break;
            case CFGMN_PROXY_PORT:
                break;
            case CFGMN_VOLUME:
                break;
            case CFGMN_NEW_IP:
                break;
            case CFGMN_NEW_PORT:
                break;
            case CFGMN_EDIT:
                break;
            case CFGMN_YES:
                break;
            }
            cpos = cfg_menu[cpos].cfgm_btn[2];
        }
    }
}

int ConfigDialogStart(void)
{
    if (NutThreadCreate("cfg", ConfigDialogThread, 0, 1024) == NULL) {
        return -1;
    }
    return 0;
}
#endif
