/*!
 * Copyright (C) 2006-2007 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * \file mp3player.c
 * \brief MP3 player.
 *
 * \verbatim
 *
 * $Log$
 *
 * \endverbatim
 */

#include <cfg/os.h>
#include <cfg/clock.h>
#include <dev/board.h>
#include <sys/heap.h>
#include <sys/timer.h>
#include <sys/event.h>

#include <stdlib.h>
#include <string.h>

#include <stdio.h>
#include <io.h>

#include "tlv320dac.h"
#include "receiver.h"
#include "mp3player.h"

static void *hres;

#define ENABLE_RESAMPLER

#ifdef ENABLE_RESAMPLER
//SAM7X static short pcmbuf[2305];
static short pcmbuf[2508];
#endif

/*!
 * \brief Wait until the MP3 data buffer is sufficiently filled.
 *
 */
static void Mp3PlayerBufferFill(PLAYERINFO *pip)
{
    RECEIVERINFO *rip;
    time_t start = time(NULL);

    printf("\nBuffering");
    for (;;) {
        /* Return if 2/3 of the buffer is filled. */
        rip = pip->pi_rcvr;
        if (rip == NULL || rip->ri_avail >= MP3_BUFSIZ / 3 * 2) {
            break;
        }
        /* Return if we reached our time limit. */
        if (time(NULL) - start >= MAX_WAIT_MP3BUF_FILLED) {
            break;
        }
        /* Return if someone stopped the player. */
        if (pip->pi_status != PSTAT_RUNNING) {
            break;
        }
        /* Wait for an event from the producer. */
        NutEventWait(&rip->ri_rdbque, 1000);
        putchar('.');
    }
    putchar('\n');
}

/*!
 *
 */
THREAD(Mp3PlayerThread, arg)
{
    PLAYERINFO *pip = (PLAYERINFO *)arg;
    RECEIVERINFO *rip;
    int rbytes;
    u_char *bufptr;
    int ec;
    int errcnt;
    int first_frame;
    int os;

    Tlv320DacInit(DAC_OUTPUT_RATE);
    Tlv320DacSetVolume(radio.rc_rvolume, radio.rc_rvolume);
    radio.rc_cvolume = radio.rc_rvolume;

    for (;;) {
        /* Idle loop. */
        for (;;) {
            printf("[PIDLE]");
            pip->pi_status = PSTAT_IDLE;
            NutEventBroadcast(&pip->pi_stsevt);

            if (pip->pi_mp3dec) {
                MP3FreeDecoder(pip->pi_mp3dec);
//                printf("\nDecoderExit %lu bytes free\n", NutHeapAvailable());
                pip->pi_mp3dec = NULL;
            }

            /* Wait for start event. */
            NutEventWait(&pip->pi_cmdevt, 0);
            printf("[PEVT%u]", pip->pi_status);
            if (pip->pi_status == PSTAT_START) {
                if (pip->pi_mp3dec == NULL) {
//                    printf("\nDecoderInit %lu bytes free\n", NutHeapAvailable());
                    if ((pip->pi_mp3dec = MP3InitDecoder()) == NULL) {
                        printf("\n[ERR-MP3INI]");
                    }
//                    printf("\nDecoderInit %lu bytes free\n", NutHeapAvailable());
                }
                if (pip->pi_mp3dec && pip->pi_rcvr) {
                    break;
                }
            }
        }

        /* Broadcast running event. */
        printf("[PRUN]");
        pip->pi_status = PSTAT_RUNNING;
        NutEventBroadcast(&pip->pi_stsevt);
        rip = pip->pi_rcvr;

        /* Wait for enough data to start playing. */
        NutSleep(500);
        Mp3PlayerBufferFill(pip);

        /* Running loop. */
        errcnt = 0;
        first_frame = 1;
        while (pip->pi_status == PSTAT_RUNNING) {
            /* Wait for new data. */
            if (rip->ri_avail < 2 * MAINBUF_SIZE) {
                Mp3PlayerBufferFill(pip);
                continue;
            }

            if (rip->ri_rpos > rip->ri_wpos && MP3_BUFSIZ - rip->ri_rpos < 2 * MAINBUF_SIZE) {
                /* The read pointer is in front of the write pointer, but
                   there is not enough MP3 data at the end of the buffer.
                   Wrap data from the front to the back, because the decoder
                   requires a continous buffer. Note, that we allocated the
                   additional bytes at the end of the MP3 data buffer for
                   this purpose. */
                memcpy(&rip->ri_buff[MP3_BUFSIZ], rip->ri_buff, 2 * MAINBUF_SIZE - (MP3_BUFSIZ - rip->ri_rpos));
            }

            /* Test for sync error. Discard data up to the next sync word. */
            ec = 0;
            if ((rbytes = MP3FindSyncWord(&rip->ri_buff[rip->ri_rpos], 2 * MAINBUF_SIZE)) < 0) {
                puts("\n[ERR-SYN]");
			    break;
		    }
            else if (rbytes == 0) {
                /* We are in sync. Decode the next MP3 frame. */
                if (ec == 0) {
                    rbytes = 2 * MAINBUF_SIZE;
                    bufptr = &rip->ri_buff[rip->ri_rpos];
 		            ec = MP3Decode(pip->pi_mp3dec, &bufptr, &rbytes, pip->pi_pcmbuf, 0);
                    if (ec) {
                        rbytes = 0;
                    }
                    else {
                        rbytes = 2 * MAINBUF_SIZE - rbytes;
//                        printf("<FRM=%d>", rbytes);
                    }
                }
                if (ec == 0) {
                    if (first_frame) {
                        MP3GetLastFrameInfo(pip->pi_mp3dec, &pip->pi_frameinfo);
                        printf("\nBitrate    : %d\n", pip->pi_frameinfo.bitrate);
                        printf("Channels   : %d\n", pip->pi_frameinfo.nChans);
                        printf("Sample rate: %d\n", pip->pi_frameinfo.samprate);
                        printf("Bits/sample: %d\n", pip->pi_frameinfo.bitsPerSample);
                        printf("Samples    : %d\n", pip->pi_frameinfo.outputSamps);
                        printf("Layer      : %d\n", pip->pi_frameinfo.layer);
                        printf("Version    : %d\n", pip->pi_frameinfo.version);
                        if (pip->pi_frameinfo.samprate > DAC_OUTPUT_RATE) {
                            printf("ERR-LIMIT %d\n", DAC_OUTPUT_RATE);
                        }
                        if (pip->pi_frameinfo.nChans == 1) {
                            pip->pi_frameinfo.samprate /= 2;
                        }
#ifdef ENABLE_RESAMPLER
                        if ((hres = RAInitResamplerHermite(pip->pi_frameinfo.samprate, DAC_OUTPUT_RATE, pip->pi_frameinfo.nChans)) == NULL) {
                            puts("\n[ERR-RESAMP]");
                            break;
                        }
#endif
                        first_frame = 0;
                    }
#ifdef ENABLE_RESAMPLER
//                    printf("<PCM=%d", pip->pi_frameinfo.outputSamps);
                    if (pip->pi_frameinfo.samprate == DAC_OUTPUT_RATE) {
                        os = pip->pi_frameinfo.outputSamps;
                        memcpy(pcmbuf, pip->pi_pcmbuf, os * 2);
                    }
                    else if (pip->pi_frameinfo.nChans == 1) {
//                        printf("[MonoMax%d]", RAGetMaxOutputHermite(pip->pi_frameinfo.outputSamps, hres));
                        os = RAResampleMonoHermite(pip->pi_pcmbuf, pip->pi_frameinfo.outputSamps, pcmbuf, hres);
                    }
                    else {
//                        printf("[StereoMax%d]", RAGetMaxOutputHermite(pip->pi_frameinfo.outputSamps, hres));
                        os = RAResampleStereoHermite(pip->pi_pcmbuf, pip->pi_frameinfo.outputSamps, pcmbuf, hres);
                    }
//                    printf("/%d>", os);
                    if (Tlv320DacWrite(pcmbuf, os)) {
                        puts("\n[**ERR-ADC**]");
                        break;
                    }
#else
                    os = pip->pi_frameinfo.outputSamps;
                    if (Tlv320DacWrite(pip->pi_pcmbuf, os)) {
                        puts("\n[**ERR-ADC**]");
                        break;
                    }
#endif
                    if (radio.rc_rvolume != radio.rc_cvolume) {
                        Tlv320DacSetVolume(radio.rc_rvolume, radio.rc_rvolume);
                        radio.rc_cvolume = radio.rc_rvolume;
                    }
                    /* Slowly recover from previous errors. */
                    if (errcnt) {
                        errcnt--;
                    }
                }
                else {
                    printf("\n[ERR%d]%lu bytes free\n", ec, NutHeapAvailable());
                    /* Move to idle state in case of fatal errors. */
                    if (ec == ERR_MP3_OUT_OF_MEMORY) {
                        break;
                    }
                    /* In case of too many errors move to idle state. */
                    if (++errcnt > MAX_PLAYERRORS) {
                        break;
                    }
                    /* Avoid to get stuck on the same error. */
                    if (rbytes == 0) {
                        rbytes = 1;
                    }
                }
            }
            else {
                /* In case of too many sync errors move to idle state. */
                printf("[SKIP %d]", rbytes);
                if (++errcnt > MAX_PLAYERRORS) {
                    break;
                }
            }
            /* Adjust the read pointer and the number of bytes available. */
            rip->ri_avail -= rbytes;
            rip->ri_rpos += rbytes;
            if (rip->ri_rpos >= MP3_BUFSIZ) {
                rip->ri_rpos -= MP3_BUFSIZ;
            }
            /* Inform the receiver, that we removed data from the buffer. */
            NutEventPost(&rip->ri_wrbque);
        }
        if (hres) {
            RAFreeResamplerHermite(hres);
            hres = NULL;
        }
        if (Tlv320DacFlush()) {
            printf("\n[ERR DACFlush]");
        }
    }
}

/*!
 * \brief Initialize MP3 player.
 */
int Mp3PlayerInit(PLAYERINFO *pip)
{
    if (pip->pi_tid == 0) {
        pip->pi_tid = NutThreadCreate("mp3play", Mp3PlayerThread, pip, 1024);
    }
    return pip->pi_tid ? 0 : -1;
}
